# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License
from __future__ import annotations

import asyncio
from contextlib import suppress
from typing import cast
from unittest.mock import call, Mock

import later
from later.unittest import TestCase
from later.unittest.mock import AsyncMock


class TaskTests(TestCase):
    async def test_as_task(self) -> None:
        tsleep = later.as_task(asyncio.sleep)
        # pep484 is still limited on typing around decorators, best to cast
        task: asyncio.Task = cast(asyncio.Task, tsleep(500))
        self.assertIsInstance(task, asyncio.Task)
        await later.cancel(task)
        self.assertTrue(task.done())
        self.assertTrue(task.cancelled())

    async def test_cancel_task(self) -> None:
        task: asyncio.Task = asyncio.get_running_loop().create_task(asyncio.sleep(500))
        await later.cancel(task)
        self.assertTrue(task.done())
        self.assertTrue(task.cancelled())

    async def test_cancel_raises_other_exception(self) -> None:
        started = False

        @later.as_task
        async def _coro():
            nonlocal started
            started = True
            try:
                await asyncio.sleep(500)
            except asyncio.CancelledError:
                raise TypeError

        task: asyncio.Task = cast(asyncio.Task, _coro())
        await asyncio.sleep(0)
        self.assertTrue(started)
        with self.assertRaises(TypeError):
            await later.cancel(task)
        self.assertTrue(task.done())
        self.assertFalse(task.cancelled())

    async def test_cancel_already_done_task(self) -> None:
        started = False

        @later.as_task
        async def _coro():
            nonlocal started
            started = True

        task: asyncio.Task = cast(asyncio.Task, _coro())
        await asyncio.sleep(0)
        self.assertTrue(started)
        self.assertTrue(task.done())
        await later.cancel(task)

    async def test_cancel_task_completes(self) -> None:
        started = False

        @later.as_task
        async def _coro():
            nonlocal started
            started = True
            try:
                await asyncio.sleep(500)
            except asyncio.CancelledError:
                return 5

        task: asyncio.Task = cast(asyncio.Task, _coro())
        await asyncio.sleep(0)
        self.assertTrue(started)
        with self.assertRaises(asyncio.InvalidStateError):
            await later.cancel(task)
        self.assertTrue(task.done())
        self.assertFalse(task.cancelled())

    async def test_cancel_when_cancelled(self) -> None:
        started, cancelled = False, False

        @later.as_task
        async def test():
            nonlocal cancelled, started
            started = True
            try:
                await asyncio.sleep(500)
            except asyncio.CancelledError:
                cancelled = True
                await asyncio.sleep(0.5)
                raise

        # neat :P
        cancel_as_task = later.as_task(later.cancel)
        otask = cast(asyncio.Task, test())  # task created a scheduled.
        await asyncio.sleep(0)  # let test start
        self.assertTrue(started)
        ctask = cast(asyncio.Task, cancel_as_task(otask))
        await asyncio.sleep(0)  # let the cancel as task start
        self.assertFalse(otask.cancelled())
        ctask.cancel()
        await asyncio.sleep(0)  # Insure the cancel was raised in the ctask
        self.assertTrue(cancelled)
        # Not done yet since the orignal task is sleeping
        self.assertFalse(ctask.cancelled())
        # we are not cancelled yet, there is a 0.5 sleep in the cancellation flow
        with self.assertRaises(asyncio.CancelledError):
            # now our cancel must raise a CancelledError as per contract
            await ctask
        self.assertTrue(ctask.cancelled())
        self.assertTrue(otask.cancelled())


class WatcherTests(TestCase):
    async def test_empty_watcher(self) -> None:
        async with later.Watcher():
            pass

    async def test_preexit_callbacks(self) -> None:
        callback = Mock()
        callback.side_effect = Exception("DERP!")

        async with later.Watcher() as w:
            w.add_preexit_callback(callback, 1, 2)

        self.assertTrue(callback.called)
        callback.assert_has_calls([call(1, 2)])

    async def test_add_task_and_remove_task(self) -> None:
        loop = asyncio.get_running_loop()

        def fixer(orig_task: asyncio.Task) -> asyncio.Task:
            return loop.create_task(asyncio.sleep(0.5))

        task: asyncio.Task = loop.create_task(asyncio.sleep(10))
        watcher = later.Watcher(context=True)
        watcher.watch(fixer=fixer)

        async def work() -> None:
            await asyncio.sleep(0.1)
            watcher.watch(task)
            await asyncio.sleep(0)
            self.assertTrue(await watcher.unwatch(task))
            self.assertTrue(await watcher.unwatch(fixer=fixer))
            await asyncio.sleep(10)

        async with watcher:
            watcher.watch(loop.create_task(work()))
            watcher.watch(task, shield=True)
            self.assertTrue(await watcher.unwatch(task, shield=True))
            self.assertFalse(await watcher.unwatch(task, shield=True))
            loop.call_later(0.2, watcher.cancel)

        task.cancel()
        with suppress(asyncio.CancelledError):
            await task

    async def test_create_task(self) -> None:
        async with later.Watcher() as w:
            w.create_task(asyncio.sleep(0.1))
            w.create_task(asyncio.sleep(0.1))
            w.create_task(asyncio.sleep(0.1))

    def test_bad_watch_call(self) -> None:
        w = later.Watcher()
        with self.assertRaises(ValueError):
            w.watch()

    async def test_watcher_fail_with_no_fix(self) -> None:
        loop = asyncio.get_running_loop()

        task: asyncio.Task = loop.create_task(asyncio.sleep(0.1))
        with self.assertRaises(RuntimeError):
            async with later.Watcher(done_ok=False) as watcher:
                watcher.watch(task)

    async def test_watcher_cancel_timeout(self) -> None:
        async def coro() -> None:
            try:
                await asyncio.sleep(300)
            except asyncio.CancelledError:
                await asyncio.sleep(2)
                raise

        loop = asyncio.get_running_loop()
        task = loop.create_task(coro())
        with self.assertRaises(later.WatcherError):
            async with later.Watcher(cancel_timeout=0.5) as watcher:
                watcher.watch(task)
                loop.call_later(0.2, watcher.cancel)
        # insure the task isn't still pending so we don't fail the later TestCase checks
        await later.cancel(task)

    async def test_watcher_cancel(self) -> None:
        loop = asyncio.get_running_loop()

        async with later.Watcher(cancel_timeout=0.5) as watcher:
            watcher.watch(loop.create_task(asyncio.sleep(500)))
            loop.call_later(0.2, watcher.cancel)

    async def test_watcher_fail_with_callable_fixer(self) -> None:
        loop = asyncio.get_running_loop()
        fixer = Mock()
        replacement_task: asyncio.Task = loop.create_task(asyncio.sleep(0.1))
        fixer.side_effect = [replacement_task, None]
        task: asyncio.Task = loop.create_task(asyncio.sleep(0.1))
        with self.assertRaises(TypeError):  # from fixer returning None
            async with later.Watcher(done_ok=False) as watcher:
                watcher.watch(task, fixer)

        self.assertTrue(fixer.called)
        fixer.assert_has_calls([call(task), call(replacement_task)])

    async def test_watcher_fail_with_async_fixer(self) -> None:
        loop = asyncio.get_running_loop()
        fixer = AsyncMock()
        replacement_task: asyncio.Task = loop.create_task(asyncio.sleep(0.1))
        fixer.side_effect = [replacement_task, None]
        task: asyncio.Task = loop.create_task(asyncio.sleep(0.1))
        with self.assertRaises(TypeError):  # from fixer returning None
            async with later.Watcher(done_ok=False) as watcher:
                watcher.watch(task, fixer)

        self.assertTrue(fixer.called)
        fixer.assert_awaited()
        fixer.assert_has_awaits([call(task), call(replacement_task)])

    async def test_watcher_START_TASK_with_bad_callable_fixer(self) -> None:
        fixer = Mock()
        fixer.return_value = None

        with self.assertRaises(TypeError):
            async with later.Watcher() as watcher:
                watcher.watch(fixer=fixer)

        self.assertTrue(fixer.called)
        fixer.assert_has_calls([call(later.START_TASK)])

    async def test_watcher_START_TASK_with_bad_async_fixer(self) -> None:
        fixer = AsyncMock()
        fixer.return_value = None

        with self.assertRaises(TypeError):
            async with later.Watcher() as watcher:
                watcher.watch(fixer=fixer)

        self.assertTrue(fixer.called)
        fixer.assert_awaited()
        fixer.assert_has_calls([call(later.START_TASK)])

    async def test_watcher_START_TASK_with_working_fixer(self) -> None:
        loop = asyncio.get_running_loop()
        fixer = AsyncMock()
        task: asyncio.Task = loop.create_task(asyncio.sleep(0.1))
        fixer.side_effect = [task, None]

        with self.assertRaises(TypeError):  # from fixer returning None
            async with later.Watcher(done_ok=False) as watcher:
                watcher.watch(fixer=fixer)

        self.assertTrue(fixer.called)
        fixer.assert_awaited()
        fixer.assert_has_awaits([call(later.START_TASK), call(task)])

    async def test_watcher_canceled_parent_aexit(self) -> None:
        loop = asyncio.get_running_loop()
        task: asyncio.Task = loop.create_task(asyncio.sleep(500))
        with self.assertRaises(asyncio.TimeoutError):
            async with later.timeout(0.2):
                async with later.Watcher() as watcher:
                    watcher.watch(task)
        self.assertTrue(task.cancelled())

    async def test_watcher_raise_in_body(self) -> None:
        loop = asyncio.get_running_loop()
        task: asyncio.Task = loop.create_task(asyncio.sleep(500))
        with self.assertRaises(RuntimeError):
            async with later.Watcher() as watcher:
                watcher.watch(task)
                await asyncio.sleep(0)
                raise RuntimeError
        self.assertTrue(task.cancelled())

    async def test_watcher_cancel_task_badly(self) -> None:
        loop = asyncio.get_running_loop()

        async def coro() -> None:
            try:
                await asyncio.sleep(300)
                return None
            except asyncio.CancelledError:
                raise Exception("OMG!")

        task: asyncio.Task = loop.create_task(coro())
        with self.assertRaises(later.WatcherError):
            async with later.Watcher(cancel_timeout=0.5) as watcher:
                watcher.watch(task)
                loop.call_later(0.2, watcher.cancel)

    def test_watch_non_task(self) -> None:
        watcher = later.Watcher()
        # Can't watch coros, or anything else
        with self.assertRaises(TypeError):
            # pyre-fixme[6]: Expected `Task[typing.Any]` for 1st param but got
            #  `Future[Variable[asyncio.tasks._T]]`.
            watcher.watch(asyncio.sleep(1))

    async def test_watcher_context(self) -> None:
        loop = asyncio.get_running_loop()

        def start_a_task():
            later.Watcher.get().watch(loop.create_task(asyncio.sleep(5)))

        async with later.Watcher() as watcher:
            loop.call_later(0.2, watcher.cancel)
            start_a_task()

    async def test_watcher_context_at_init(self) -> None:
        loop = asyncio.get_running_loop()

        def start_a_task():
            later.Watcher.get().watch(loop.create_task(asyncio.sleep(5)))

        watcher = later.Watcher(context=True)
        start_a_task()
        async with watcher:
            loop.call_later(0.2, watcher.cancel)

    def test_watcher_context_non_exist(self) -> None:
        with self.assertRaises(LookupError):
            later.Watcher.get()

    async def test_watch_with_shield(self) -> None:
        loop = asyncio.get_running_loop()
        task: asyncio.Task[None] = loop.create_task(asyncio.sleep(0.1))
        async with later.Watcher() as watcher:
            watcher.watch(task, shield=True)
            with self.assertRaises(ValueError):
                # This is not permitted
                watcher.watch(task, fixer=lambda x: task, shield=True)
            watcher.cancel()

        await task


class HerdTests(TestCase):
    async def test_herd_cancellation(self) -> None:
        called = 0
        original_cancelled = False

        @later.herd
        # pyre-fixme[3]: Return type must be annotated.
        # pyre-fixme[2]: Parameter must be annotated.
        async def fun(event):
            nonlocal called
            nonlocal original_cancelled
            called += 1
            try:
                await event.wait()
            except asyncio.CancelledError:
                original_cancelled = True
                raise

        event = asyncio.Event()
        call1 = asyncio.create_task(fun(event))
        call2 = asyncio.create_task(fun(event))

        done, pending = await asyncio.wait({call1, call2}, timeout=0.05)
        self.assertFalse(done)
        self.assertEqual(called, 1)
        self.assertFalse(original_cancelled)

        await later.cancel(call1)
        self.assertFalse(original_cancelled)

        # even with the first call cancelled call2 can continue
        done, pending = await asyncio.wait([call2], timeout=0.05)
        self.assertFalse(done)

        # Only after there is only one pending left do we allow the original task
        # to be cancelled.
        await later.cancel(call2)
        self.assertTrue(original_cancelled)

    async def test_herd(self) -> None:
        called = 0
        waited = 0

        @later.herd(ignored_args={1})
        # pyre-fixme[3]: Return type must be annotated.
        # pyre-fixme[2]: Parameter must be annotated.
        async def fun(event, ignored_arg):
            nonlocal called
            called += 1
            await event.wait()
            nonlocal waited
            waited += 1

        event = asyncio.Event()
        call1 = asyncio.create_task(fun(event, object()))
        call2 = asyncio.create_task(fun(event, object()))
        call3 = asyncio.create_task(fun(event, object()))

        done, pending = await asyncio.wait({call1, call2, call3}, timeout=0.05)
        self.assertFalse(done)
        self.assertEqual(called, 1)
        self.assertEqual(waited, 0)

        event.set()
        done, pending = await asyncio.wait({call1, call2, call3})
        self.assertTrue(done)
        self.assertFalse(pending)
        self.assertEqual(called, 1)
        self.assertEqual(waited, 1)

    async def test_herd_exception(self) -> None:
        called = 0

        @later.herd
        # pyre-fixme[3]: Return type must be annotated.
        # pyre-fixme[2]: Parameter must be annotated.
        async def fun(event):
            nonlocal called
            called += 1
            await event.wait()
            raise RuntimeError

        event = asyncio.Event()
        call1 = asyncio.create_task(fun(event))
        call2 = asyncio.create_task(fun(event))

        done, pending = await asyncio.wait({call1, call2}, timeout=0.05)
        self.assertFalse(done)
        self.assertEqual(called, 1)

        event.set()
        done, pending = await asyncio.wait({call1, call2})
        self.assertTrue(done)
        self.assertFalse(pending)
        self.assertEqual(called, 1)
        # access the exception so we don't blow up the test
        with self.assertRaises(RuntimeError):
            await call1

        with self.assertRaises(RuntimeError):
            await call2

    def test_herd_usage_no_args(self) -> None:
        @later.herd
        async def hax(test: int) -> None:
            """docstring"""
            ...

        # Testing wraps
        self.assertEqual(hax.__name__, "hax")
        self.assertEqual(hax.__doc__, "docstring")



# This file was copied from python 3.8 stdlib, for use in python 3.7
# It is licensed as python see https://github.com/python/cpython/blob/master/LICENSE
# Any changes made for 3.7 were done under that license.

import asyncio
import inspect
import sys
from unittest.case import _Outcome, TestCase as _TestCase


class TestCase(_TestCase):  # pragma: nocover not testing the non async pieces
    def run(self, result=None):
        orig_result = result
        if result is None:
            result = self.defaultTestResult()
            startTestRun = getattr(result, "startTestRun", None)
            if startTestRun is not None:
                startTestRun()
        result.startTest(self)
        testMethod = getattr(self, self._testMethodName)
        if getattr(self.__class__, "__unittest_skip__", False) or getattr(
            testMethod, "__unittest_skip__", False
        ):
            # If the class or method was skipped.
            try:
                skip_why = getattr(
                    self.__class__, "__unittest_skip_why__", ""
                ) or getattr(testMethod, "__unittest_skip_why__", "")
                self._addSkip(result, self, skip_why)
            finally:
                result.stopTest(self)
            return
        expecting_failure_method = getattr(
            testMethod, "__unittest_expecting_failure__", False
        )
        expecting_failure_class = getattr(self, "__unittest_expecting_failure__", False)
        expecting_failure = expecting_failure_class or expecting_failure_method
        outcome = _Outcome(result)
        try:
            self._outcome = outcome

            with outcome.testPartExecutor(self):
                self._callSetUp()
            if outcome.success:
                outcome.expecting_failure = expecting_failure
                with outcome.testPartExecutor(self, isTest=True):
                    self._callTestMethod(testMethod)
                outcome.expecting_failure = False
                with outcome.testPartExecutor(self):
                    self._callTearDown()

            self.doCleanups()
            for test, reason in outcome.skipped:
                self._addSkip(result, test, reason)
            self._feedErrorsToResult(result, outcome.errors)
            if outcome.success:
                if expecting_failure:
                    if outcome.expectedFailure:
                        self._addExpectedFailure(result, outcome.expectedFailure)
                    else:
                        self._addUnexpectedSuccess(result)
                else:
                    result.addSuccess(self)
            return result
        finally:
            result.stopTest(self)
            if orig_result is None:
                stopTestRun = getattr(result, "stopTestRun", None)
                if stopTestRun is not None:
                    stopTestRun()
            # explicitly break reference cycles:
            # outcome.errors -> frame -> outcome -> outcome.errors
            # outcome.expectedFailure -> frame -> outcome -> outcome.expectedFailure
            outcome.errors.clear()
            outcome.expectedFailure = None
            # clear the outcome, no more needed
            self._outcome = None

    def doCleanups(self):
        """Execute all cleanup functions. Normally called for you after
        tearDown."""
        outcome = self._outcome or _Outcome()
        while self._cleanups:
            function, args, kwargs = self._cleanups.pop()
            with outcome.testPartExecutor(self):
                self._callCleanup(function, *args, **kwargs)

        # return this for backwards compatibility
        # even though we no longer use it internally
        return outcome.success


if sys.version_info[:2] < (3, 8):

    def handle_base_exception(fut, exc):
        # Due to a bug in 3.7, that was fixed in 3.8, if any BaseException is raised
        # from from a test will make it hang (discussion https://fburl.com/5ul4eol2).
        # This workaround sets up a new Exception that tries to mimic the original
        # BaseException name / message.
        exc_class = type(type(exc).__name__, (Exception,), {})
        new_ex = exc_class(str(exc))
        new_ex.__cause__ = exc
        fut.set_exception(new_ex)

else:

    def handle_base_exception(fut, exc):
        fut.set_exception(exc)


class IsolatedAsyncioTestCase(TestCase):
    # Names intentionally have a long prefix
    # to reduce a chance of clashing with user-defined attributes
    # from inherited test case
    #
    # The class doesn't call loop.run_until_complete(self.setUp()) and family
    # but uses a different approach:
    # 1. create a long-running task that reads self.setUp()
    #    awaitable from queue along with a future
    # 2. await the awaitable object passing in and set the result
    #    into the future object
    # 3. Outer code puts the awaitable and the future object into a queue
    #    with waiting for the future
    # The trick is necessary because every run_until_complete() call
    # creates a new task with embedded ContextVar context.
    # To share contextvars between setUp(), test and tearDown() we need to execute
    # them inside the same task.

    # Note: the test case modifies event loop policy if the policy was not instantiated
    # yet.
    # asyncio.get_event_loop_policy() creates a default policy on demand but never
    # returns None
    # I believe this is not an issue in user level tests but python itself for testing
    # should reset a policy in every test module
    # by calling asyncio.set_event_loop_policy(None) in tearDownModule()

    def __init__(self, methodName="runTest"):
        super().__init__(methodName)
        self._asyncioTestLoop = None
        self._asyncioCallsQueue = None

    async def asyncSetUp(self):
        pass

    async def asyncTearDown(self):
        pass

    def addAsyncCleanup(self, func, *args, **kwargs):
        # A trivial trampoline to addCleanup()
        # the function exists because it has a different semantics
        # and signature:
        # addCleanup() accepts regular functions
        # but addAsyncCleanup() accepts coroutines
        #
        # We intentionally don't add inspect.iscoroutinefunction() check
        # for func argument because there is no way
        # to check for async function reliably:
        # 1. It can be "async def func()" iself
        # 2. Class can implement "async def __call__()" method
        # 3. Regular "def func()" that returns awaitable object
        self.addCleanup(*(func, *args), **kwargs)

    def _callSetUp(self):
        self.setUp()
        self._callAsync(self.asyncSetUp)

    def _callTestMethod(self, method):
        self._callMaybeAsync(method)

    def _callTearDown(self):
        self._callAsync(self.asyncTearDown)
        self.tearDown()

    def _callCleanup(self, function, *args, **kwargs):
        self._callMaybeAsync(function, *args, **kwargs)

    def _callAsync(self, func, *args, **kwargs):
        assert self._asyncioTestLoop is not None
        ret = func(*args, **kwargs)
        assert inspect.isawaitable(ret)
        fut = self._asyncioTestLoop.create_future()
        self._asyncioCallsQueue.put_nowait((fut, ret))
        return self._asyncioTestLoop.run_until_complete(fut)

    def _callMaybeAsync(self, func, *args, **kwargs):
        assert self._asyncioTestLoop is not None
        ret = func(*args, **kwargs)
        if inspect.isawaitable(ret):
            fut = self._asyncioTestLoop.create_future()
            self._asyncioCallsQueue.put_nowait((fut, ret))
            return self._asyncioTestLoop.run_until_complete(fut)
        else:
            return ret  # pragma: nocover

    async def _asyncioLoopRunner(self, fut):
        self._asyncioCallsQueue = queue = asyncio.Queue()
        fut.set_result(None)
        while True:
            query = await queue.get()
            queue.task_done()
            if query is None:
                return
            fut, awaitable = query
            try:
                ret = await awaitable
                if not fut.cancelled():
                    fut.set_result(ret)
            except asyncio.CancelledError:  # pragma: nocover
                raise
            except Exception as ex:
                if not fut.cancelled():
                    fut.set_exception(ex)
            except BaseException as ex:
                if not fut.cancelled():
                    handle_base_exception(fut, ex)

    def _setupAsyncioLoop(self):
        assert self._asyncioTestLoop is None
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.set_debug(True)
        self._asyncioTestLoop = loop
        fut = loop.create_future()
        self._asyncioCallsTask = loop.create_task(self._asyncioLoopRunner(fut))
        loop.run_until_complete(fut)

    def _tearDownAsyncioLoop(self):
        assert self._asyncioTestLoop is not None
        loop = self._asyncioTestLoop
        self._asyncioTestLoop = None
        self._asyncioCallsQueue.put_nowait(None)
        loop.run_until_complete(self._asyncioCallsQueue.join())

        try:
            # cancel all tasks
            to_cancel = asyncio.all_tasks(loop)
            if not to_cancel:
                return

            for task in to_cancel:
                task.cancel()

            if sys.version_info >= (3, 10):
                loop.run_until_complete(
                    asyncio.gather(*to_cancel, return_exceptions=True)
                )
            else:
                loop.run_until_complete(
                    asyncio.gather(*to_cancel, loop=loop, return_exceptions=True)
                )

            for task in to_cancel:
                if task.cancelled():
                    continue
                if task.exception() is not None:  # pragma: nocover
                    loop.call_exception_handler(
                        {
                            "message": "unhandled exception during test shutdown",
                            "exception": task.exception(),
                            "task": task,
                        }
                    )
            # shutdown asyncgens
            loop.run_until_complete(loop.shutdown_asyncgens())
        finally:
            asyncio.set_event_loop(None)
            loop.close()

    def run(self, result=None):
        self._setupAsyncioLoop()
        try:
            return super().run(result)
        finally:
            self._tearDownAsyncioLoop()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# This file was copied from python 3.8 test suite, for use in python 3.7
# It is licensed as python itself see https://github.com/python/cpython/LICENSE
# Any changes made for 3.7 were done under that license.

import asyncio
import unittest

from later.unittest.backport.async_case import IsolatedAsyncioTestCase


def tearDownModule() -> None:
    asyncio.set_event_loop_policy(None)


class TestHangsForever(IsolatedAsyncioTestCase):
    @unittest.expectedFailure
    async def test_base_exception_from_async_method(self):
        raise BaseException("Hangs forever")


class TestAsyncCase(unittest.TestCase):
    def test_full_cycle(self) -> None:
        events = []

        class Test(IsolatedAsyncioTestCase):
            def setUp(self):
                self.assertEqual(events, [])
                events.append("setUp")

            async def asyncSetUp(self):
                self.assertEqual(events, ["setUp"])
                events.append("asyncSetUp")

            async def test_func(self):
                self.assertEqual(events, ["setUp", "asyncSetUp"])
                events.append("test")
                self.addAsyncCleanup(self.on_cleanup)

            async def asyncTearDown(self):
                self.assertEqual(events, ["setUp", "asyncSetUp", "test"])
                events.append("asyncTearDown")

            def tearDown(self):
                self.assertEqual(
                    events, ["setUp", "asyncSetUp", "test", "asyncTearDown"]
                )
                events.append("tearDown")

            async def on_cleanup(self):
                self.assertEqual(
                    events, ["setUp", "asyncSetUp", "test", "asyncTearDown", "tearDown"]
                )
                events.append("cleanup")

        test = Test("test_func")
        test.run()
        self.assertEqual(
            events,
            ["setUp", "asyncSetUp", "test", "asyncTearDown", "tearDown", "cleanup"],
        )

    def test_exception_in_setup(self) -> None:
        events = []

        class Test(IsolatedAsyncioTestCase):
            async def asyncSetUp(self):
                events.append("asyncSetUp")
                raise Exception()

            async def test_func(self):
                events.append("test")
                self.addAsyncCleanup(self.on_cleanup)

            async def asyncTearDown(self):
                events.append("asyncTearDown")

            async def on_cleanup(self):
                events.append("cleanup")

        test = Test("test_func")
        test.run()
        self.assertEqual(events, ["asyncSetUp"])

    def test_exception_in_test(self) -> None:
        events = []

        class Test(IsolatedAsyncioTestCase):
            async def asyncSetUp(self):
                events.append("asyncSetUp")

            async def test_func(self):
                events.append("test")
                raise Exception()
                self.addAsyncCleanup(self.on_cleanup)

            async def asyncTearDown(self):
                events.append("asyncTearDown")

            async def on_cleanup(self):
                events.append("cleanup")

        test = Test("test_func")
        test.run()
        self.assertEqual(events, ["asyncSetUp", "test", "asyncTearDown"])

    def test_exception_in_test_after_adding_cleanup(self) -> None:
        events = []

        class Test(IsolatedAsyncioTestCase):
            async def asyncSetUp(self):
                events.append("asyncSetUp")

            async def test_func(self):
                events.append("test")
                self.addAsyncCleanup(self.on_cleanup)
                raise Exception()

            async def asyncTearDown(self):
                events.append("asyncTearDown")

            async def on_cleanup(self):
                events.append("cleanup")

        test = Test("test_func")
        test.run()
        self.assertEqual(events, ["asyncSetUp", "test", "asyncTearDown", "cleanup"])

    def test_exception_in_tear_down(self) -> None:
        events = []

        class Test(IsolatedAsyncioTestCase):
            async def asyncSetUp(self):
                events.append("asyncSetUp")

            async def test_func(self):
                events.append("test")
                self.addAsyncCleanup(self.on_cleanup)

            async def asyncTearDown(self):
                events.append("asyncTearDown")
                raise Exception()

            async def on_cleanup(self):
                events.append("cleanup")

        test = Test("test_func")
        test.run()
        self.assertEqual(events, ["asyncSetUp", "test", "asyncTearDown", "cleanup"])

    def test_exception_in_tear_clean_up(self) -> None:
        events = []

        class Test(IsolatedAsyncioTestCase):
            async def asyncSetUp(self):
                events.append("asyncSetUp")

            async def test_func(self):
                events.append("test")
                self.addAsyncCleanup(self.on_cleanup)

            async def asyncTearDown(self):
                events.append("asyncTearDown")

            async def on_cleanup(self):
                events.append("cleanup")
                raise Exception()

        test = Test("test_func")
        test.run()
        self.assertEqual(events, ["asyncSetUp", "test", "asyncTearDown", "cleanup"])

    def test_cleanups_interleave_order(self) -> None:
        events = []

        class Test(IsolatedAsyncioTestCase):
            async def test_func(self):
                self.addAsyncCleanup(self.on_sync_cleanup, 1)
                self.addAsyncCleanup(self.on_async_cleanup, 2)
                self.addAsyncCleanup(self.on_sync_cleanup, 3)
                self.addAsyncCleanup(self.on_async_cleanup, 4)

            async def on_sync_cleanup(self, val):
                events.append(f"sync_cleanup {val}")

            async def on_async_cleanup(self, val):
                events.append(f"async_cleanup {val}")

        test = Test("test_func")
        test.run()
        self.assertEqual(
            events,
            ["async_cleanup 4", "sync_cleanup 3", "async_cleanup 2", "sync_cleanup 1"],
        )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""
This TestCase attempts to track all tasks so that they are ensured to have
been awaited. Any time asyncio calls logger.error() it is considered a
test failure.
"""
from __future__ import annotations

import asyncio
import asyncio.base_tasks
import asyncio.coroutines
import asyncio.futures
import asyncio.log
import asyncio.tasks
import reprlib
import sys
import unittest.mock as mock
import weakref
from functools import wraps
from typing import Any, Callable, TypeVar


if sys.version_info[:2] >= (3, 9):  # pragma: nocover
    from unittest import IsolatedAsyncioTestCase as AsyncioTestCase
else:  # pragma: nocover
    from .backport.async_case import IsolatedAsyncioTestCase as AsyncioTestCase


_F = TypeVar("_F", bound=Callable[..., Any])
_IGNORE_TASK_LEAKS_ATTR = "__later_testcase_ignore_tasks__"
_IGNORE_AIO_ERRS_ATTR = "__later_testcase_ignore_asyncio__"
atleastpy38: bool = sys.version_info[:2] >= (3, 8)
_unmanaged_tasks: weakref.WeakSet[asyncio.Task] = weakref.WeakSet()


class TestTask(asyncio.Task):
    _managed: bool = False
    _coro_repr: str

    def __init__(self, coro, *args, **kws) -> None:
        # pyre-fixme[16]: Module `coroutines` has no attribute `_format_coroutine`.
        self._coro_repr = asyncio.coroutines._format_coroutine(coro)
        _unmanaged_tasks.add(self)
        super().__init__(coro, *args, **kws)

    @reprlib.recursive_repr()
    def __repr__(self) -> str:
        # pyre-fixme[16]: `TestTask` has no attribute `_repr_info`.
        repr_info = asyncio.base_tasks._task_repr_info(self)
        coro = f"coro={self._coro_repr}"
        if atleastpy38:  # py3.8 added name=
            repr_info[2] = coro  # pragma: nocover
        else:
            repr_info[1] = coro  # pragma: nocover
        return f"<{self.__class__.__name__} {' '.join(repr_info)}>"

    def _mark_managed(self):
        self._managed = True

    def __await__(self):
        self._mark_managed()
        return super().__await__()

    def result(self):
        if self.done():
            self._mark_managed()
        return super().result()

    def exception(self):
        if self.done():
            self._mark_managed()
        return super().exception()

    def add_done_callback(self, fn, *, context=None) -> None:
        @wraps(fn)
        def mark_managed(fut):
            self._mark_managed()
            return fn(fut)

        super().add_done_callback(mark_managed, context=context)

    def was_managed(self) -> bool:
        if self._managed:
            return True
        # If the task is done() and the result is None, let it pass as managed
        # We use super here so we don't manage ourselves.
        return (
            self.done()
            and not self.cancelled()
            and not super().exception()
            and super().result() is None
        )

    def __del__(self) -> None:
        # So a pattern is to create_task, and not save the results.
        # we accept that as long as there was no result other than None
        # thrift-py3 uses this pattern to call rpc methods in ServiceInterfaces
        # where any result/execption is returned to the remote client.
        if not self.was_managed():
            context = {
                "task": self,
                "message": (
                    "Task was destroyed but never awaited!, "
                    f"WrappedCoro: {self._coro_repr}"
                ),
            }
            # pyre-fixme[16]: `TestTask` has no attribute `_source_traceback`.
            if self._source_traceback:
                context["source_traceback"] = self._source_traceback
            self._loop.call_exception_handler(context)
        super().__del__()


def task_factory(loop, coro, **kws) -> TestTask:
    task = TestTask(coro, loop=loop, **kws)
    return task


def all_tasks(loop):
    # This is a copy of asyncio.all_task but returns even done tasks
    # so we can see if they were awaited instead of ignored

    # This is copied from the guts of asyncio.all_tasks
    # Looping the WeakSet since it is possible it fails during iteration
    _get_loop = asyncio.futures._get_loop
    i = 0
    while True:
        try:
            tasks = list(_unmanaged_tasks)
        except RuntimeError:  # pragma: nocover
            i += 1
            if i >= 1000:
                raise
        else:
            break
    return {t for t in tasks if _get_loop(t) is loop}


def ignoreAsyncioErrors(test_item: _F) -> _F:
    """Test is allowed to cause Asyncio Error Logs"""
    setattr(test_item, _IGNORE_AIO_ERRS_ATTR, True)
    return test_item


def ignoreTaskLeaks(test_item: _F) -> _F:
    """Test is allowed to leak tasks"""
    setattr(test_item, _IGNORE_TASK_LEAKS_ATTR, True)
    return test_item


class TestCase(AsyncioTestCase):
    def _callTestMethod(self, testMethod) -> None:
        ignore_error = getattr(
            self,
            _IGNORE_AIO_ERRS_ATTR,
            getattr(testMethod, _IGNORE_AIO_ERRS_ATTR, False),
        )
        ignore_tasks = getattr(
            self,
            _IGNORE_TASK_LEAKS_ATTR,
            getattr(testMethod, _IGNORE_TASK_LEAKS_ATTR, False),
        )
        if sys.version_info >= (3, 11):  # pragma: nocover
            loop = self._asyncioRunner.get_loop()
        else:
            # pyre-fixme[16]: `TestCase` has no attribute `_asyncioTestLoop`.
            loop = self._asyncioTestLoop
        if not ignore_tasks:
            # install our own task factory for monitoring usage
            loop.set_task_factory(task_factory)

        # Track existing tasks
        start_tasks = all_tasks(loop)
        # Setup a patch for the asyncio logger
        real_logger = asyncio.log.logger.error
        with mock.patch.object(
            asyncio.log.logger, "error", side_effect=real_logger
        ) as error:
            # pyre-fixme[16]: `AsyncioTestCase` has no attribute `_callTestMethod`.
            super()._callTestMethod(testMethod)

        if sys.version_info < (3, 11):
            # Lets join the queue to insure all the tasks created by this case
            # are cleaned up
            # pyre-fixme[16]: `TestCase` has no attribute `_asyncioCallsQueue`.
            loop.run_until_complete(self._asyncioCallsQueue.join())
        left_over_tasks = set(all_tasks(loop)) - set(start_tasks)
        for task in list(left_over_tasks):
            if isinstance(task, TestTask) and task.was_managed():
                left_over_tasks.remove(task)
        if left_over_tasks and not ignore_tasks:
            self.assertEqual(set(), left_over_tasks, "left over un-awaited tasks!")
        if error.called and not ignore_error:
            errors = "\n\n".join(c[0][0] for c in error.call_args_list)
            self.fail(f"asyncio logger.error() was called!\n{errors}")



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License
from __future__ import annotations

import asyncio
import unittest
from typing import Any, List

from later.unittest import ignoreAsyncioErrors, ignoreTaskLeaks, TestCase


# This is a place to purposefully produce leaked tasks
saved_tasks: List[asyncio.Task[Any]] = []


class TestTestCase(TestCase):
    @unittest.expectedFailure
    async def test_unmanaged_task_created_in_testmethod(self) -> None:
        saved_tasks.append(asyncio.get_running_loop().create_task(asyncio.sleep(10)))

    async def test_managed_task_done(self) -> None:
        task = asyncio.get_running_loop().create_task(asyncio.sleep(0.5, "test"))
        saved_tasks.append(task)
        await task

    async def test_unmanaged_task_done_none(self) -> None:
        """None value Done tasks are ignored"""

        async def coro(e: asyncio.Event) -> None:
            e.set()

        event = asyncio.Event()
        saved_tasks.append(asyncio.get_running_loop().create_task(coro(event)))
        await event.wait()

    @unittest.expectedFailure
    async def test_unmanaged_task_done_value(self) -> None:
        async def coro(e: asyncio.Event) -> bool:
            e.set()
            return False

        event = asyncio.Event()
        saved_tasks.append(asyncio.get_running_loop().create_task(coro(event)))
        await event.wait()

    @unittest.expectedFailure
    async def test_garbage_collected_task_during_testmethod(self) -> None:
        t = asyncio.get_running_loop().create_task(asyncio.sleep(10))
        t.cancel()
        del t

    async def test_tasks_add_callback(self) -> None:
        # If we add a callback and it is run then thats good enough
        # to say the task was managed
        t = asyncio.get_running_loop().create_task(asyncio.sleep(10))
        x = False

        def callback(fut):
            nonlocal x
            x = True

        t.add_done_callback(callback)
        t.cancel()
        await asyncio.sleep(0.1)  # callbacks are scheduled
        self.assertTrue(x, "callback was executed")

    async def test_task_exception(self) -> None:
        async def coro():
            raise RuntimeError

        t = asyncio.get_running_loop().create_task(coro())
        # Letting the task move to done, without awaiting it
        while not t.done():
            await asyncio.sleep(0)
        t.exception()
        # pyre-fixme[16]: `Task` has no attribute `was_managed`.
        self.assertTrue(t.was_managed())
        self.assertIsInstance(t.exception(), RuntimeError)
        self.assertTrue(t.done())

    @ignoreAsyncioErrors
    async def test_ignore_asyncio_error(self) -> None:
        async def sub():
            return 5

        sub()  # don't await

    @ignoreTaskLeaks
    async def test_ignore_task_leaks(self) -> None:
        async def coro():
            raise RuntimeError

        saved_tasks.append(asyncio.get_running_loop().create_task(coro()))

    async def test_forgotten_tasks_done_no_value(self) -> None:
        asyncio.get_running_loop().create_task(asyncio.sleep(0))
        await asyncio.sleep(0)


@ignoreAsyncioErrors
class IgnoreAsyncioErrorsTestCase(TestCase):
    async def test_ignore_asyncio_error_on_case_class(self) -> None:
        async def sub():
            return 5

        sub()  # don't await


@ignoreTaskLeaks
class IgnoreTaskLeaksTestCase(TestCase):
    async def test_ignore_task_leaks_on_case_class(self) -> None:
        async def coro():
            raise RuntimeError

        saved_tasks.append(asyncio.get_running_loop().create_task(coro()))


if __name__ == "__main__":
    unittest.main()



# This file was copied from typeshed and modified to expose types for IsolatedAsyncioTestCase
# It is licensed as typeshed see https://github.com/python/typeshed/blob/master/LICENSE
import sys
from typing import Any, Awaitable, Callable
from unittest import TestCase

class IsolatedAsyncioTestCase(TestCase):
    async def asyncSetUp(self) -> None: ...
    async def asyncTearDown(self) -> None: ...
    def addAsyncCleanup(
        self, __func: Callable[..., Awaitable[Any]], *args: Any, **kwargs: Any
    ) -> None: ...



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys
import unittest

from later.tests.test_event import Test_BiDirectionalEvent  # noqa: F401
from later.tests.test_task import TaskTests, WatcherTests  # noqa: F401
from later.tests.test_version import VersionTests  # noqa: F401

if sys.version_info[:2] < (3, 9):
    from later.tests.unittest.backport.test_async_case import (  # noqa: F401
        TestAsyncCase,
        TestHangsForever,
    )
    from later.tests.unittest.backport.test_mock import TestPatch  # noqa: F401
from later.tests.unittest.test_case import (  # noqa: F401
    IgnoreAsyncioErrorsTestCase,
    IgnoreTaskLeaksTestCase,
    TestTestCase,
)
from later.tests.unittest.test_mock import TestAsyncContextManagerMock  # noqa: F401


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
import contextvars
import functools
import logging
import threading

from collections.abc import Coroutine, Generator
from functools import partial, wraps
from inspect import isawaitable
from types import TracebackType
from typing import (
    AbstractSet,
    Any,
    Awaitable,
    Callable,
    cast,
    Dict,
    Hashable,
    List,
    Mapping,
    NewType,
    Optional,
    overload,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from unittest.mock import Mock

from .event import BiDirectionalEvent


FixerType = Callable[[asyncio.Task], Union[asyncio.Task, Awaitable[asyncio.Task]]]
logger = logging.getLogger(__name__)
T = TypeVar("T")
F = TypeVar("F", bound=Callable[..., Awaitable[Any]])

__all__: Sequence[str] = [
    "Watcher",
    "START_TASK",
    "TaskSentinel",
    "cancel",
    "as_task",
    "herd",
]


class TaskSentinel(asyncio.Task):
    """When you need a done task for typing"""

    def __init__(self):
        fake = Mock()
        asyncio.Future.__init__(self, loop=fake)  # typing: ignore, don't create a loop
        asyncio.Future.set_result(self, None)


async def cancel(fut: asyncio.Future) -> None:
    """
    Cancel a future/task and await for it to cancel.
    If the fut is already done() this is a no-op
    If everything goes well this returns None.

    If this coroutine is cancelled, we wait for the passed in argument to cancel
    but we will raise the CancelledError as per Cancellation Contract, Unless the task
    doesn't cancel correctly then we could raise other exceptions.

    If the task raises an exception during cancellation we re-raise it
    if the task completes instead of cancelling we raise a InvalidStateError
    """
    if fut.done():
        return  # nothing to do
    fut.cancel()
    exc: Optional[asyncio.CancelledError] = None
    while not fut.done():
        shielded = asyncio.shield(fut)
        try:
            await asyncio.wait([shielded])
        except asyncio.CancelledError as ex:
            exc = ex
        finally:
            # Insure we handle the exception/value that may exist on the shielded task
            # This will prevent errors logged to the asyncio logger
            if (
                shielded.done()
                and not shielded.cancelled()
                and not shielded.exception()
            ):
                shielded.result()
    if fut.cancelled():
        if exc is None:
            return
        # we were cancelled also so honor the contract
        raise exc from None
    # Some exception thrown during cancellation
    ex = fut.exception()
    if ex is not None:
        raise ex from None
    # fut finished instead of cancelled, wat?
    raise asyncio.InvalidStateError(
        f"task didn't raise CancelledError on cancel: {fut} had result {fut.result()}"
    )


def as_task(func: F) -> F:
    """
    Decorate a function, So that when called it is wrapped in a task
    on the running loop.
    """

    @wraps(func)
    def create_task(*args, **kws):
        loop = asyncio.get_running_loop()
        return loop.create_task(func(*args, **kws))

    return cast(F, create_task)


# Sentinel Task
START_TASK: asyncio.Task = TaskSentinel()

# ContextVar for Finding an existing Task Watcher
WATCHER_CONTEXT: contextvars.ContextVar[Watcher] = contextvars.ContextVar(
    "WATCHER_CONTEXT"
)


class WatcherError(RuntimeError):
    pass


# pyre-fixme[13]: Attribute `loop` is never initialized.
class Watcher:
    _tasks: Dict[asyncio.Future, Optional[FixerType]]
    _scheduled: List[FixerType]
    _tasks_changed: BiDirectionalEvent
    _cancelled: asyncio.Event
    _cancel_timeout: float
    _preexit_callbacks: List[Callable[[], None]]
    _shielded_tasks: Dict[asyncio.Task, asyncio.Future]
    loop: asyncio.AbstractEventLoop
    running: bool
    done_ok: bool

    @staticmethod
    def get() -> Watcher:
        return WATCHER_CONTEXT.get()

    def __init__(
        self,
        *,
        cancel_timeout: float = 300,
        context: bool = False,
        done_ok: bool = True,
    ) -> None:
        """
        cancel_timeout is the time in seconds we will wait after cancelling all
        the tasks watched by this watcher.

        context is wether to expose this Watcher via contextvars now or at __aenter__
        """
        if context:
            WATCHER_CONTEXT.set(self)
        self._cancel_timeout = cancel_timeout
        self._tasks = {}
        self._scheduled = []
        self._tasks_changed = BiDirectionalEvent()
        self._cancelled = asyncio.Event()
        self._preexit_callbacks = []
        self._shielded_tasks = {}
        self.running = False
        self.done_ok = done_ok

    async def _run_scheduled(self) -> None:
        scheduled = self._scheduled
        while scheduled:
            fixer = scheduled.pop()
            task = fixer(START_TASK)
            if not isinstance(task, asyncio.Task) and isawaitable(task):
                task = await task

            if isinstance(task, asyncio.Task):
                self._tasks[task] = fixer
            else:
                raise TypeError(f"{fixer}(START_TASK) failed to return a task.")

    async def unwatch(
        self,
        task: asyncio.Task = START_TASK,
        fixer: Optional[FixerType] = None,
        *,
        shield: bool = False,
    ) -> bool:
        """
        The ability to unwatch a task, by task or fixer
        This is a coroutine to insure the watcher has re-watched the tasks list

        If the task was shielded then you need to specify here so we can find
        the shield and remove it from the watch list.

        When unwatching a fixer, if the returned task is not the same
        as the one passed in we will cancel it, and await it.
        """

        async def tasks_changed():
            if self.running:
                await self._tasks_changed.set()

        if shield:
            if task in self._shielded_tasks:
                del self._tasks[self._shielded_tasks[task]]
                del self._shielded_tasks[task]
                await tasks_changed()
                return True
        elif fixer is not None:
            for t, fix in tuple(self._tasks.items()):
                if fix is fixer:
                    del self._tasks[t]
                    await tasks_changed()
                    if t is not task:
                        await cancel(t)
                    return True
        elif task is not START_TASK:
            if task in self._tasks:
                del self._tasks[task]
                await tasks_changed()
                return True
        return False

    def watch(
        self,
        task: asyncio.Task = START_TASK,
        fixer: Optional[FixerType] = None,
        *,
        shield: bool = False,
    ) -> None:
        """
        Add a task to be watched by the watcher
        You can also attach a fixer co-routine or function to be used to fix a
        task that has died.

        The fixer will be passed the failed task, and is expected to return a working
        task, or raise if that is impossible.

        You can also just pass in the fixer and we will use it to create the task
        to be watched.  The fixer will be passed a dummy task singleton:
        `later.task.START_TASK`

        shield argument lets you watch a task, but not cancel it in this watcher.
        Useful for triggering on task failures, but not managing said task.
        """
        # Watching a coro, leads to a confusing error deep in watcher
        # so use runtime checks not just static types.
        if not isinstance(task, asyncio.Task):
            raise TypeError("only asyncio.Task objects can be watched.")

        if task is START_TASK:
            if not fixer:
                raise ValueError("fixer must be specified when using START_TASK.")
            self._scheduled.append(fixer)
        elif shield:
            if fixer:
                raise ValueError("`fixer` can not be used with shield=True")
            self._shielded_tasks[task] = asyncio.shield(task)
            self._tasks[self._shielded_tasks[task]] = None
        else:
            self._tasks[task] = fixer
        self._tasks_changed.set_nowait()

    def create_task(
        self, coro: Generator[Any, None, T] | Coroutine[Any, Any, T], **kws
    ) -> asyncio.Task[T]:
        t = asyncio.create_task(coro, **kws)
        self.watch(t)
        return t

    def cancel(self) -> None:
        """
        Stop the watcher and cause it to cancel all the tasks in its care.
        """
        self._cancelled.set()

    def add_preexit_callback(self, callback: Callable[..., None], *args, **kws) -> None:
        self._preexit_callbacks.append(partial(callback, *args, **kws))

    def _run_preexit_callbacks(self) -> None:
        for callback in self._preexit_callbacks:
            try:
                callback()
            except Exception as e:
                logger.exception(
                    f"ignoring exception from pre-exit callback {callback}: {e}"
                )

    async def __aenter__(self) -> "Watcher":
        WATCHER_CONTEXT.set(self)
        self.loop = asyncio.get_running_loop()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> bool:
        cancel_task: asyncio.Task = self.loop.create_task(self._cancelled.wait())
        changed_task: asyncio.Task = START_TASK
        try:
            # an exception was raised in the body of the watcher.
            # just return, cancel any tasks.
            if exc is not None:
                return False
            self.running = True
            while not self._cancelled.is_set():
                if self._scheduled:
                    await self._run_scheduled()
                if changed_task is START_TASK or changed_task.done():
                    changed_task = self.loop.create_task(self._tasks_changed.wait())
                if not self._tasks:
                    return False  # There are no tasks just exit.
                done, pending = await asyncio.wait(
                    [cancel_task, changed_task, *self._tasks.keys()],
                    return_when=asyncio.FIRST_COMPLETED,
                )
                if cancel_task in done:
                    break  # Don't bother doing fixes just break out
                for task in done:
                    task = cast(asyncio.Task, task)
                    if task is changed_task:
                        continue
                    else:
                        await self._fix_task(task)
        finally:
            self.running = False
            self._run_preexit_callbacks()
            await self._event_task_cleanup(cancel_task, changed_task)
            await self._handle_cancel()
            self._tasks.clear()
            self._shielded_tasks.clear()
        return False

    async def _event_task_cleanup(self, *tasks):
        for task in tasks:
            if task is not START_TASK:
                await cancel(task)

    async def _fix_task(self, task: asyncio.Task) -> None:
        # Insure we "retrieve" the result of failed tasks
        exc = task.exception()
        if exc is None:
            task.result()
            if self.done_ok:
                # clean up the task is done. And thats okay.
                del self._tasks[task]
                return
        fixer = self._tasks[task]
        if fixer is None:
            raise RuntimeError(f"{task} finished and there is no fixer!") from exc
        new_task = fixer(task)
        if not isinstance(new_task, asyncio.Task) and isawaitable(new_task):
            new_task = await new_task

        if isinstance(new_task, asyncio.Task):
            del self._tasks[task]
            self._tasks[new_task] = fixer
        else:
            raise TypeError(
                f"{fixer}(task) failed to return a task, returned:" f"{new_task}!"
            ) from exc

    async def _handle_cancel(self):
        tasks = [task for task in self._tasks if not task.done()]
        if not tasks:
            return

        for task in tasks:
            task.cancel()

        done, pending = await asyncio.wait(tasks, timeout=self._cancel_timeout)
        bad_tasks: List[asyncio.Task] = []
        for task in done:
            if task.cancelled():
                continue
            if task.exception() is not None:
                bad_tasks.append(task)

        bad_tasks.extend(pending)

        if bad_tasks:
            raise WatcherError(
                "The following tasks didn't cancel cleanly or at all!", bad_tasks
            )


CacheKey = NewType("CacheKey", Sequence[Hashable])
ArgID = Union[int, str]


class _CountTask:
    """So herd can track herd size and task together for cancellation"""

    task: Optional[asyncio.Task] = None
    count: int = 0


def _get_local(local: threading.local, field: str) -> Dict[CacheKey, object]:
    """
    helper for attempting to fetch a named attr from a threading.local
    """
    try:
        return cast(Dict[CacheKey, object], getattr(local, field))
    except AttributeError:
        container: Dict[CacheKey, object] = {}
        setattr(local, field, container)
        return container


def _build_key(
    args: Tuple[object, ...],
    kwargs: Mapping[str, object],
    ignored_args: Optional[AbstractSet[ArgID]] = None,
) -> CacheKey:
    """
    Build a key for caching Hashable args and kwargs.
    Allow for not including certain fields from args or kwargs
    """
    if not ignored_args:
        # pyre-fixme[45]: Cannot instantiate abstract class `CacheKey`.
        return CacheKey((args, tuple(sorted(kwargs.items()))))

    # If we do want to ignore something then do so
    # pyre-fixme[45]: Cannot instantiate abstract class `CacheKey`.
    return CacheKey(
        (
            tuple((value for idx, value in enumerate(args) if idx not in ignored_args)),
            tuple(
                (item for item in sorted(kwargs.items()) if item[0] not in ignored_args)
            ),
        )
    )


@overload  # noqa: 811
def herd(
    fn: F, *, ignored_args: Optional[AbstractSet[ArgID]] = None
) -> F:  # pragma: nocover
    ...


@overload  # noqa: 811
def herd(
    fn: Optional[F] = None, *, ignored_args: Optional[AbstractSet[ArgID]] = None
) -> Callable[[F], F]:  # pragma: nocover
    ...


def herd(
    fn=None,
    *,
    ignored_args: Optional[AbstractSet[ArgID]] = None,
):  # noqa: 811
    """
    Provide a simple thundering herd protection as a decorator.
    if requests comes in while and existing request with those same args is pending,
    wait for the pending request and return its results.

    ignored_args are arguments that should be ignored for matching with
    existing requests. Use arg position or kwargs name.
    Example: a client arg for when multiple clients exists but the request hits the same
    backend.

    Each member of the herd is "shielded" from cancellation effecting other herd members
    """

    def decorator(fn: F) -> F:
        local = threading.local()

        @functools.wraps(fn)
        async def wrapped(*args, **kwargs):
            pending = cast(Dict[CacheKey, _CountTask], _get_local(local, "pending"))
            request = _build_key(tuple(args), kwargs, ignored_args)
            count_task = pending.setdefault(request, _CountTask())
            count_task.count += 1
            task = count_task.task  # thanks pyre
            if task is None:
                count_task.task = task = asyncio.create_task(fn(*args, **kwargs))
            try:
                return await asyncio.shield(task)
            except asyncio.CancelledError:
                if count_task.count == 1:
                    await cancel(task)
                raise  # always re-raise CancelledError
            finally:
                count_task.count -= 1
                # Lets destroy the herd on last member exit or
                # First success member exit. This is to mirror the original
                # herd behavior that tore down the herd after the original call exited
                if count_task.count == 0 or not task.cancelled():
                    if request in pending and pending[request] is count_task:
                        del pending[request]

        return cast(F, wrapped)

    if fn and callable(fn):
        # pyre-fixme[6]: For 1st param expected `F` but got `(...) -> object`.
        return decorator(fn)

    return decorator



# fb has some local typeshed for unittest.mock with more than ANY for everything
# So we rely on what ever typehed is available to the end user.

from typing import Any, List, Optional, Sequence
from unittest.mock import (
    ANY as ANY,
    Base,
    call as call,
    create_autospec as create_autospec,
    DEFAULT as DEFAULT,
    FILTER_DIR as FILTER_DIR,
    MagicMock as MagicMock,
    Mock as Mock,
    mock_open as mock_open,
    NonCallableMagicMock as NonCallableMagicMock,
    NonCallableMock as NonCallableMock,
    patch as patch,
    PropertyMock as PropertyMock,
    seal as seal,
    sentinel as sentinel,
)

class AsyncMockMixin(Base):
    await_count: int
    # pyre-fixme[11]: Annotation `call` is not defined as a type.
    await_args: Optional[call]
    await_args_list: List[call]
    def assert_awaited(self) -> None: ...
    def assert_awaited_once(self) -> None: ...
    def assert_awaited_with(self, *args: Any, **kws: Any) -> None: ...
    def assert_awaited_once_with(self, *args: Any, **kws: Any) -> None: ...
    def assert_any_await(self, *args: Any, **kws: Any) -> None: ...
    def assert_has_awaits(
        self, calls: Sequence[call], any_order: bool = ...
    ) -> None: ...
    def assert_not_awaited(self) -> None: ...

class AsyncMock(AsyncMockMixin, Mock): ...



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""A toolbox for asyncio services"""

from async_timeout import timeout

from .event import BiDirectionalEvent
from .task import as_task, cancel, herd, START_TASK, Watcher, WatcherError


__version__ = "23.11.30"
__all__ = [
    "BiDirectionalEvent",
    "START_TASK",
    "Watcher",
    "WatcherError",
    "cancel",
    "timeout",
    "as_task",
    "herd",
]



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from asyncio import Event


class BiDirectionalEvent:
    """
    Events can be set and cleared multiple times, but this one has backchannel
    information for the setter to ensure the waiter has called wait() a second time.
    Both the wait and the set are coroutines
    """

    _sevent: Event
    _cevent: Event
    _first_wait: Event

    def __init__(self) -> None:
        self._cevent = Event()
        self._sevent = Event()
        self._first_wait = Event()

    async def wait(self) -> None:
        if self._first_wait.is_set():
            self._cevent.set()
        else:
            self._first_wait.set()
        await self._sevent.wait()
        self._sevent.clear()

    def is_set(self) -> bool:
        return self._sevent.is_set()

    async def set(self) -> None:
        if not self._first_wait.is_set():
            await self._first_wait.wait()
        self._sevent.set()
        await self._cevent.wait()
        self._cevent.clear()

    def set_nowait(self) -> None:
        self._sevent.set()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License
import asyncio
from contextlib import suppress

from later import BiDirectionalEvent
from later.unittest import TestCase


class Test_BiDirectionalEvent(TestCase):
    async def test_end_to_end(self) -> None:
        event = BiDirectionalEvent()
        woke = False

        async def waiter():
            nonlocal woke
            while True:
                # do some work
                await event.wait()
                woke = True

        task = asyncio.get_running_loop().create_task(waiter())
        self.assertFalse(woke)
        await event.set()
        self.assertTrue(woke)
        woke = False
        await event.set()
        self.assertTrue(woke)
        self.assertFalse(event.is_set())
        event.set_nowait()
        self.assertTrue(event.is_set())
        # clean up the task
        task.cancel()
        with suppress(asyncio.CancelledError):
            await task



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import sys

from typing import Any
from unittest.mock import call, DEFAULT

if sys.version_info >= (3, 9):  # pragma: nocover
    from unittest.mock import AsyncMock, create_autospec, MagicMock, Mock, patch
else:  # pragma: nocover
    from .backport.mock import AsyncMock, create_autospec, MagicMock, Mock, patch


__all__ = [
    "AsyncContextManager",
    "AsyncMock",
    "MagicMock",
    "Mock",
    "call",
    "create_autospec",
    "patch",
]


def AsyncContextManager(return_value: Any = DEFAULT, instance: bool = False) -> Mock:
    """
    This helper sets up a Mock Tree to mock out an AsyncContextManager class.
    return_value: The object returned by the context manager
    instance: Instead of a Class, this mock represents and instance


    mock.assert_awaited() -> The context manager was __aentered__
    mock.assert_not_awaited() -> the context manager was not __aentered__
    """
    mock = Mock()
    manager_instance = AsyncMock()

    if instance:
        mock = manager_instance
    else:
        mock.return_value = manager_instance

    if return_value is not DEFAULT:
        manager_instance.__aenter__.return_value = return_value

    mock.assert_awaited = manager_instance.__aenter__.assert_awaited
    mock.assert_not_awaited = manager_instance.__aenter__.assert_not_awaited
    return mock



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License
from unittest.mock import call

from later.unittest import TestCase
from later.unittest.mock import AsyncContextManager


class TestAsyncContextManagerMock(TestCase):
    async def test_return_value(self) -> None:

        mock = AsyncContextManager(return_value=10)

        async with mock() as val:
            self.assertEqual(val, 10)
            mock.assert_awaited()
            mock.assert_called()

    async def test_call_args(self) -> None:
        c = call(expire=True)
        mock = AsyncContextManager()

        async with mock(expire=True):
            mock.assert_awaited()
            self.assertEqual(mock.call_args, c)

    async def test_as_instance(self) -> None:
        mock = AsyncContextManager(return_value=10, instance=True)

        async with mock as val:
            mock.assert_awaited()
            self.assertEqual(val, 10)

    async def test_created_but_not_aentered(self) -> None:

        mock = AsyncContextManager()

        m = mock()

        mock.assert_called()
        mock.assert_not_awaited()

        async with m:
            pass

        mock.assert_awaited()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License
import unittest

from later.unittest.backport.async_case import IsolatedAsyncioTestCase
from later.unittest.backport.mock import patch


class Helper:
    @classmethod
    async def async_class_method(cls):
        raise Exception("Should never happen")

    @staticmethod
    async def async_static_method():
        raise Exception("Should never happen")

    async def async_method(self):
        raise Exception("Should never happen")


class TestPatch(IsolatedAsyncioTestCase):
    @patch("later.tests.unittest.backport.test_mock.Helper.async_class_method")
    async def test_patch_classmethods(self, patched) -> None:
        await Helper.async_class_method()

    @patch("later.tests.unittest.backport.test_mock.Helper.async_static_method")
    async def test_patch_staticmethods(self, patched) -> None:
        await Helper.async_static_method()

    @patch("later.tests.unittest.backport.test_mock.Helper.async_method")
    async def test_patch_methods(self, patched) -> None:
        await Helper().async_method()


if __name__ == "__main__":
    unittest.main()
